#include "fdbuf.h"
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

///////////////////////////////////////////////////////////////////////////////
// Class fdbuf
///////////////////////////////////////////////////////////////////////////////
fdbuf::fdbuf(int fdesc, bool dc, unsigned bufsize)
  : buf(new char[bufsize]),
    length(0),
    start(0),
    errnum(0),
    flags(0),
    size(bufsize),
    fd(fdesc),
    do_close(dc)
{
  if(!buf) {
    flags = flag_error;
    errnum = errno;
  }
  if(fdesc < 0)
    flags |= flag_closed;
}

fdbuf::~fdbuf()
{
  close();
}

bool fdbuf::error() const
{
  return flags & flag_error;
}

bool fdbuf::closed() const
{
  return flags & flag_closed;
}

bool fdbuf::close()
{
  if(do_close && fd >= 0)
    if(::close(fd) == -1) {
      errnum = errno;
      flags |= flag_error;
      return false;
    }
  flags |= flag_closed;
  return true;
}

///////////////////////////////////////////////////////////////////////////////
// Class fdibuf
///////////////////////////////////////////////////////////////////////////////
fdibuf::fdibuf(int fdesc, bool dc, unsigned bufsize)
  : fdbuf(fdesc, dc, bufsize)
{
}

fdibuf::fdibuf(const char* filename, unsigned bufsize)
  : fdbuf(open(filename, O_RDONLY), true, bufsize)
{
  if(fd == -1) {
    flags = flag_error;
    errnum = errno;
  }
}

fdibuf::~fdibuf()
{
}

bool fdibuf::eof() const
{
  return (flags & flag_eof) && (start >= length);
}

bool fdibuf::operator!() const
{
  return eof() || error() || closed();
}

bool fdibuf::refill()
{
  if(flags)
    return false;
  if(start != 0) {
    if(start < length) {
      length -= start;
      memcpy(buf, buf+start, length);
    } else
      length = 0;
    start = 0;
  }
  unsigned oldlength = length;
  if(length < size) {
    ssize_t red = ::read(fd, buf+length, size-length);
    if(red == -1) {
      errnum = errno;
      flags |= flag_error;
    }
    else if(red == 0)
      flags |= flag_eof;
    else
      length += red;
  }
  return length > oldlength;
}

bool fdibuf::getline(mystring& out, char terminator)
{
  if(start >= length)
    refill();
  if(eof() || error())
    return false;
  out = "";
  while(!eof() && !error()) {
    char* ptr = buf+start;
    const char* end = (const char*)memchr(ptr, terminator, length-start);
    if(!end) {
      out += mystring(ptr, length-start);
      start = length;
      refill();
    }
    else {
      out += mystring(ptr, end-ptr);
      start += (end-ptr)+1;
      break;
    }
  }
  return true;
}

bool fdibuf::read(char* data, unsigned datalen)
{
  char* ptr = data;
  count = 0;
  while(datalen && !eof()) {
    if(start >= length)
      refill();
    unsigned len = length-start;
    if(len > datalen)
      len = datalen;
    memcpy(ptr, buf+start, len);
    start += len;
    datalen -= len;
    ptr += len;
    count += len;
  }
  return !datalen;
}

///////////////////////////////////////////////////////////////////////////////
// Class fdobuf
///////////////////////////////////////////////////////////////////////////////
fdobuf::fdobuf(int fdesc, bool dc, unsigned bufsize)
  : fdbuf(fdesc, dc, bufsize)
{
}

fdobuf::fdobuf(const char* filename, fdobuf::openflags f, unsigned bufsize)
  : fdbuf(open(filename, O_WRONLY | f, 0666), true, bufsize)
{
  if(fd == -1) {
    flags = flag_error;
    errnum = errno;
  }
}

fdobuf::~fdobuf()
{
  flush();
}

bool fdobuf::operator!() const
{
  return error() || closed();
}

bool fdobuf::flush()
{
  if(flags)
    return false;
  while(start < length) {
    ssize_t written = ::write(fd, buf+start, length-start);
    if(written == -1) {
      flags |= flag_error;
      errnum = errno;
      return false;
    }
    else
      start += written;
  }
  length = 0;
  start = 0;
  return true;
}

bool fdobuf::write(const char* data, unsigned datalen)
{
  const char* ptr = data;
  count = 0;
  while(datalen) {
    unsigned amount = size-length;
    if(amount > datalen)
      amount = datalen;
    memcpy(buf+length, ptr, amount);
    length += amount;
    datalen -= amount;
    ptr += amount;
    if(length >= size && !flush())
      return false;
    count += amount;
  }
  return true;
}

///////////////////////////////////////////////////////////////////////////////
// Other routines
///////////////////////////////////////////////////////////////////////////////
bool fdbuf_copy(fdibuf& in, fdobuf& out)
{
  if(!in || !out)
    return false;
  while(!in.eof()) {
    char buf[BUFSIZE];
    if(!in.read(buf, BUFSIZE) && in.last_count() == 0)
      return false;
    if(!out.write(buf, in.last_count()) && out.last_count() < in.last_count())
      return false;
  }
  if(!out.flush())
    return false;
  return true;
}
